@@ -99,7 +99,7 @@ class Jp2k(Jp2kBox):
         cbsize=None, cinema2k=None, cinema4k=None, colorspace=None,
         cratios=None, eph=None, grid_offset=None, irreversible=None, mct=None,
         modesw=None, numres=None, plt=False, prog=None, psizes=None, psnr=None,
-        sop=None, subsam=None,
+        sop=None, subsam=None, tlm=False
     ):
         """
         Parameters
@@ -156,6 +156,8 @@ class Jp2k(Jp2kBox):
             Subsampling factors (dy, dx).
         tilesize : tuple, optional
             Tile size in terms of (numrows, numcols), not (X, Y).
+        tlm : bool, optional
+            Generate TLM markers.
         verbose : bool, optional
             Print informational messages produced by the OpenJPEG library.
         """
@@ -189,8 +191,11 @@ class Jp2k(Jp2kBox):
         self._sop = sop
         self._subsam = subsam
         self._tilesize = tilesize
+        self._tlm = tlm
 
         self._shape = None
+        self._ndim = None
+        self._dtype = None
         self._ignore_pclr_cmap_cdef = False
         self._verbose = verbose
 
@@ -369,6 +374,45 @@ class Jp2k(Jp2kBox):
 
         self._layer = layer
 
+    @property
+    def dtype(self):
+        """
+        Datatype of the image elements.
+        """
+        if self._dtype is None:
+            c = self.get_codestream()
+            bps0 = c.segment[1].bitdepth[0]
+            sgnd0 = c.segment[1].signed[0]
+
+            if (
+                all(bitdepth == bps0 for bitdepth in c.segment[1].bitdepth)
+                and all(signed == sgnd0 for signed in c.segment[1].signed)
+            ):
+                if bps0 <= 8:
+                    self._dtype = np.int8 if sgnd0 else np.uint8
+                else:
+                    self._dtype = np.int16 if sgnd0 else np.uint16
+            else:
+                msg = (
+                    "The dtype property is only valid when all components "
+                    "have the same bitdepth and sign. "
+                    "\n\n"
+                    f"{c.segment[1]}"
+                )
+                raise TypeError(msg)
+
+        return self._dtype
+
+    @property
+    def ndim(self):
+        """
+        Number of image dimensions.
+        """
+        if self._ndim is None:
+            self._ndim = len(self.shape)
+
+        return self._ndim
+
     @property
     def codestream(self):
         if self._codestream is None:
@@ -858,6 +902,9 @@ class Jp2k(Jp2kBox):
             if self._plt:
                 opj2.encoder_set_extra_options(codec, plt=self._plt)
 
+            if self._tlm:
+                opj2.encoder_set_extra_options(codec, tlm=self._tlm)
+
             strm = opj2.stream_create_default_file_stream(self.filename, False)
 
             num_threads = get_option('lib.num_threads')
