@@ -20,12 +20,6 @@ import warnings
 # Third party library imports ...
 from lxml import etree as ET
 import numpy as np
-try:
-    import skimage.data
-    import skimage.metrics
-    _HAVE_SCIKIT_IMAGE = True
-except ModuleNotFoundError:
-    _HAVE_SCIKIT_IMAGE = False
 
 # Local imports
 import glymur
@@ -49,8 +43,9 @@ def load_tests(loader, tests, ignore):  # pragma: no cover
         # Can't do it on windows, temporary file issue.
         return tests
     if glymur.lib.openjp2.OPENJP2 is not None:
-        tests.addTests(doctest.DocTestSuite('glymur.jp2k',
-                                            tearDown=docTearDown))
+        tests.addTests(
+            doctest.DocTestSuite('glymur.jp2k', tearDown=docTearDown)
+        )
     return tests
 
 
@@ -64,6 +59,74 @@ class TestJp2k(fixtures.TestCommon):
         super(TestJp2k, self).setUp()
         glymur.reset_option('all')
 
+    def test_dtype_jp2(self):
+        """
+        Scenario:  An RGB image is read from a JP2 file.
+
+        Expected response:  the dtype property is np.uint8
+        """
+        j = Jp2k(self.jp2file)
+        self.assertEqual(j.dtype, np.uint8)
+
+    def test_dtype_j2k_uint16(self):
+        """
+        Scenario:  A uint16 monochrome image is read from a J2K file.
+
+        Expected response:  the dtype property is np.uint16
+        """
+        with ir.path('tests.data', 'uint16.j2k') as path:
+            j = Jp2k(path)
+        self.assertEqual(j.dtype, np.uint16)
+
+    def test_dtype_prec4_signd1(self):
+        """
+        Scenario:  A 4-bit signed image is read from a J2k file.
+
+        Expected response:  the dtype property is np.int8
+        """
+        with ir.path('tests.data', 'p0_03.j2k') as path:
+            j = Jp2k(path)
+        self.assertEqual(j.dtype, np.int8)
+
+    def test_dtype_inconsistent_bitdetph(self):
+        """
+        Scenario:  The image has different bitdepths in different components.
+
+        Expected response:  TypeError when accessing the dtype property.
+        """
+        with ir.path('tests.data', 'issue982.j2k') as path:
+            j = Jp2k(path)
+        with self.assertRaises(TypeError):
+            j.dtype
+
+    def test_ndims_jp2(self):
+        """
+        Scenario:  An RGB image is read from a JP2 file.
+
+        Expected response:  the ndim attribute/property is 3
+        """
+        j = Jp2k(self.jp2file)
+        self.assertEqual(j.ndim, 3)
+
+    def test_ndims_j2k(self):
+        """
+        Scenario:  An RGB image is read from a raw codestream.
+
+        Expected response:  the ndim attribute/property is 3
+        """
+        j = Jp2k(self.j2kfile)
+        self.assertEqual(j.ndim, 3)
+
+    def test_ndims_monochrome_j2k(self):
+        """
+        Scenario:  An monochrome image is read from a raw codestream.
+
+        Expected response:  the ndim attribute/property is 2
+        """
+        with ir.path('tests.data', 'p0_02.j2k') as path:
+            j = Jp2k(path)
+        self.assertEqual(j.ndim, 2)
+
     def test_read_bands_unequal_subsampling(self):
         """
         SCENARIO:  The read_bands method is used on an image with unequal
@@ -152,8 +215,9 @@ class TestJp2k(fixtures.TestCommon):
 
         channel_type = [COLOR, COLOR, COLOR]
         association = [BLUE, GREEN, RED]
-        cdef = glymur.jp2box.ChannelDefinitionBox(channel_type=channel_type,
-                                                  association=association)
+        cdef = glymur.jp2box.ChannelDefinitionBox(
+            channel_type=channel_type, association=association
+        )
         jp2h.box.append(cdef)
 
         boxes.append(jp2h)
@@ -254,8 +318,10 @@ class TestJp2k(fixtures.TestCommon):
         j2 = Jp2k(self.temp_j2k_filename, data=expdata, irreversible=True)
 
         codestream = j2.get_codestream()
-        self.assertEqual(codestream.segment[2].xform,
-                         glymur.core.WAVELET_XFORM_9X7_IRREVERSIBLE)
+        self.assertEqual(
+            codestream.segment[2].xform,
+            glymur.core.WAVELET_XFORM_9X7_IRREVERSIBLE
+        )
 
         actdata = j2[:]
         self.assertTrue(fixtures.mse(actdata[0], expdata[0]) < 0.38)
@@ -302,6 +368,9 @@ class TestJp2k(fixtures.TestCommon):
         with self.assertRaises(FileNotFoundError):
             j[:]
 
+    @unittest.skipIf(
+        not fixtures.HAVE_SCIKIT_IMAGE, fixtures.HAVE_SCIKIT_IMAGE_MSG
+    )
     def test_write_to_a_file_using_context_manager(self):
         """
         SCENARIO:  Write to a file using a context manager, read the data back
@@ -309,7 +378,7 @@ class TestJp2k(fixtures.TestCommon):
 
         EXPECTED RESULT: the data matches
         """
-        expected = skimage.data.astronaut()
+        expected = fixtures.skimage.data.astronaut()
 
         j1 = Jp2k(self.temp_jp2_filename)
         j1[:] = expected
@@ -605,18 +674,20 @@ class TestJp2k(fixtures.TestCommon):
                 # Write the ULST box.
                 # Length is 26, 1 UUID, hard code that UUID as zeros.
                 ulst_len = 26
-                write_buffer = struct.pack('>I4sHIIII', ulst_len, b'ulst',
-                                           int(1), int(0), int(0), int(0),
-                                           int(0))
+                write_buffer = struct.pack(
+                    '>I4sHIIII',
+                    ulst_len, b'ulst', int(1), int(0), int(0), int(0), int(0)
+                )
                 tfile.write(write_buffer)
 
                 # Write the URL box.
                 # Length is 16, version is one byte, flag is 3 bytes, url
                 # is the rest.
                 url_box_len = 16
-                write_buffer = struct.pack('>I4sBBBB',
-                                           url_box_len, b'url ',
-                                           int(0), int(0), int(0), int(0))
+                write_buffer = struct.pack(
+                    '>I4sBBBB',
+                    url_box_len, b'url ', int(0), int(0), int(0), int(0)
+                )
                 tfile.write(write_buffer)
                 write_buffer = struct.pack('>ssss', b'a', b'b', b'c', b'd')
                 tfile.write(write_buffer)
@@ -840,9 +911,11 @@ class TestJp2k(fixtures.TestCommon):
                           'color media'))
         self.assertEqual(profile['Rendering Intent'], 'perceptual')
 
-        np.testing.assert_almost_equal(profile['Illuminant'],
-                                       (0.964203, 1.000000, 0.824905),
-                                       decimal=6)
+        np.testing.assert_almost_equal(
+            profile['Illuminant'],
+            (0.964203, 1.000000, 0.824905),
+            decimal=6
+        )
 
         self.assertEqual(profile['Creator'], 'JPEG')
 
@@ -1073,7 +1146,9 @@ class TestJp2k_write(fixtures.MetadataBase):
             with self.assertRaises(InvalidJp2kError):
                 Jp2k(tfile.name, data=np.zeros((0, 256), dtype=np.uint8))
 
-    @unittest.skipIf(not _HAVE_SCIKIT_IMAGE, "No scikit-image found")
+    @unittest.skipIf(
+        not fixtures.HAVE_SCIKIT_IMAGE, fixtures.HAVE_SCIKIT_IMAGE_MSG
+    )
     def test_psnr_zero_value_not_last(self):
         """
         SCENARIO:  The PSNR keyword argument has a zero value, but it is not
@@ -1082,12 +1157,57 @@ class TestJp2k_write(fixtures.MetadataBase):
         EXPECTED RESULT:  RuntimeError
         """
         kwargs = {
-            'data': skimage.data.camera(),
+            'data': fixtures.skimage.data.camera(),
             'psnr': [0, 35, 40, 30],
         }
         with self.assertRaises(RuntimeError):
             Jp2k(self.temp_jp2_filename, **kwargs)
 
+    @unittest.skipIf(glymur.version.openjpeg_version < '2.5.0',
+                     "Requires as least v2.5.0")
+    def test_tlm_yes(self):
+        """
+        SCENARIO:  Use the tlm keyword.
+
+        EXPECTED RESULT:  A TLM segment is detected.
+        """
+        kwargs = {
+            'data': self.jp2_data,
+            'tlm': True
+        }
+        j = Jp2k(self.temp_jp2_filename, **kwargs)
+
+        codestream = j.get_codestream(header_only=False)
+
+        at_least_one_tlm_segment = any(
+            isinstance(seg, glymur.codestream.TLMsegment)
+            for seg in codestream.segment
+        )
+        self.assertTrue(at_least_one_tlm_segment)
+
+    def test_tlm_no(self):
+        """
+        SCENARIO:  Use the tlm keyword set to False
+
+        EXPECTED RESULT:  A TLM segment not detected.
+        """
+        kwargs = {
+            'data': self.jp2_data,
+            'tlm': False
+        }
+        j = Jp2k(self.temp_jp2_filename, **kwargs)
+
+        codestream = j.get_codestream(header_only=False)
+
+        at_least_one_tlm_segment = any(
+            isinstance(seg, glymur.codestream.TLMsegment)
+            for seg in codestream.segment
+        )
+        self.assertFalse(at_least_one_tlm_segment)
+
+    @unittest.skipIf(
+        not fixtures.HAVE_SCIKIT_IMAGE, fixtures.HAVE_SCIKIT_IMAGE_MSG
+    )
     def test_plt_yes(self):
         """
         SCENARIO:  Use the plt keyword.
@@ -1095,7 +1215,7 @@ class TestJp2k_write(fixtures.MetadataBase):
         EXPECTED RESULT:  Plt segment is detected.
         """
         kwargs = {
-            'data': skimage.data.camera(),
+            'data': fixtures.skimage.data.camera(),
             'plt': True
         }
         j = Jp2k(self.temp_jp2_filename, **kwargs)
@@ -1108,6 +1228,9 @@ class TestJp2k_write(fixtures.MetadataBase):
         )
         self.assertTrue(at_least_one_plt)
 
+    @unittest.skipIf(
+        not fixtures.HAVE_SCIKIT_IMAGE, fixtures.HAVE_SCIKIT_IMAGE_MSG
+    )
     def test_plt_no(self):
         """
         SCENARIO:  Use the plt keyword set to false.
@@ -1115,7 +1238,7 @@ class TestJp2k_write(fixtures.MetadataBase):
         EXPECTED RESULT:  Plt segment is not detected.
         """
         kwargs = {
-            'data': skimage.data.camera(),
+            'data': fixtures.skimage.data.camera(),
             'plt': False
         }
         j = Jp2k(self.temp_jp2_filename, **kwargs)
@@ -1128,7 +1251,9 @@ class TestJp2k_write(fixtures.MetadataBase):
         )
         self.assertFalse(at_least_one_plt)
 
-    @unittest.skipIf(not _HAVE_SCIKIT_IMAGE, "No scikit-image found")
+    @unittest.skipIf(
+        not fixtures.HAVE_SCIKIT_IMAGE, fixtures.HAVE_SCIKIT_IMAGE_MSG
+    )
     def test_psnr_non_zero_non_monotonically_decreasing(self):
         """
         SCENARIO:  The PSNR keyword argument is non-monotonically increasing
@@ -1137,13 +1262,15 @@ class TestJp2k_write(fixtures.MetadataBase):
         EXPECTED RESULT:  RuntimeError
         """
         kwargs = {
-            'data': skimage.data.camera(),
+            'data': fixtures.skimage.data.camera(),
             'psnr': [30, 35, 40, 30],
         }
         with self.assertRaises(RuntimeError):
             Jp2k(self.temp_jp2_filename, **kwargs)
 
-    @unittest.skipIf(not _HAVE_SCIKIT_IMAGE, "No scikit-image found")
+    @unittest.skipIf(
+        not fixtures.HAVE_SCIKIT_IMAGE, fixtures.HAVE_SCIKIT_IMAGE_MSG
+    )
     def test_psnr(self):
         """
         SCENARIO:  Four peak signal-to-noise ratio values are supplied, the
@@ -1152,7 +1279,7 @@ class TestJp2k_write(fixtures.MetadataBase):
         EXPECTED RESULT:  Four quality layers, the first should be lossless.
         """
         kwargs = {
-            'data': skimage.data.camera(),
+            'data': fixtures.skimage.data.camera(),
             'psnr': [30, 35, 40, 0],
         }
         with open(self.temp_jp2_filename, mode='wb') as tfile:
@@ -1168,8 +1295,8 @@ class TestJp2k_write(fixtures.MetadataBase):
             # warning
             warnings.simplefilter('ignore')
             psnr = [
-                skimage.metrics.peak_signal_noise_ratio(
-                    skimage.data.camera(), d[j]
+                fixtures.skimage.metrics.peak_signal_noise_ratio(
+                    fixtures.skimage.data.camera(), d[j]
                 )
                 for j in range(4)
             ]
@@ -1252,8 +1379,12 @@ class TestJp2k_write(fixtures.MetadataBase):
         size are present.  The precinct sizes validate.
         """
         with open(self.temp_j2k_filename, mode='wb') as tfile:
-            j = Jp2k(tfile.name, data=self.jp2_data, psnr=[30, 35, 40],
-                     cbsize=(16, 16), psizes=[(64, 64)])
+            j = Jp2k(
+                tfile.name,
+                data=self.jp2_data,
+                psnr=[30, 35, 40],
+                cbsize=(16, 16), psizes=[(64, 64)]
+            )
 
             codestream = j.get_codestream()
 
@@ -1266,14 +1397,16 @@ class TestJp2k_write(fixtures.MetadataBase):
         self.assertEqual(codestream.segment[2].num_res, 5)  # levels
         self.assertEqual(tuple(codestream.segment[2].code_block_size),
                          (16, 16))  # cblksz
-        self.verify_codeblock_style(codestream.segment[2].cstyle,
-                                    [False, False,
-                                     False, False, False, False])
+        self.verify_codeblock_style(
+            codestream.segment[2].cstyle,
+            [False, False, False, False, False, False]
+        )
         self.assertEqual(codestream.segment[2].xform,
                          glymur.core.WAVELET_XFORM_5X3_REVERSIBLE)
-        self.assertEqual(codestream.segment[2].precinct_size,
-                         ((2, 2), (4, 4), (8, 8), (16, 16), (32, 32),
-                          (64, 64)))
+        self.assertEqual(
+            codestream.segment[2].precinct_size,
+            ((2, 2), (4, 4), (8, 8), (16, 16), (32, 32), (64, 64))
+        )
 
     def test_NR_ENC_Bretagne2_ppm_4_encode(self):
         """
@@ -1283,12 +1416,14 @@ class TestJp2k_write(fixtures.MetadataBase):
 
         """
         with open(self.temp_j2k_filename, mode='wb') as tfile:
-            j = Jp2k(tfile.name,
-                     data=self.jp2_data,
-                     psizes=[(128, 128)] * 3,
-                     cratios=[100, 20, 2],
-                     tilesize=(480, 640),
-                     cbsize=(32, 32))
+            j = Jp2k(
+                tfile.name,
+                data=self.jp2_data,
+                psizes=[(128, 128)] * 3,
+                cratios=[100, 20, 2],
+                tilesize=(480, 640),
+                cbsize=(32, 32)
+            )
 
             # Should be three layers.
             codestream = j.get_codestream()
@@ -1305,16 +1440,25 @@ class TestJp2k_write(fixtures.MetadataBase):
             self.assertEqual(codestream.segment[2].layers, 3)  # layers = 3
             self.assertEqual(codestream.segment[2].mct, 1)  # mct
             self.assertEqual(codestream.segment[2].num_res, 5)  # levels
-            self.assertEqual(tuple(codestream.segment[2].code_block_size),
-                             (32, 32))  # cblksz
-            self.verify_codeblock_style(codestream.segment[2].cstyle,
-                                        [False, False,
-                                         False, False, False, False])
-            self.assertEqual(codestream.segment[2].xform,
-                             glymur.core.WAVELET_XFORM_5X3_REVERSIBLE)
-            self.assertEqual(codestream.segment[2].precinct_size,
-                             ((16, 16), (32, 32), (64, 64), (128, 128),
-                              (128, 128), (128, 128)))
+            self.assertEqual(
+                tuple(codestream.segment[2].code_block_size),
+                (32, 32)
+            )  # cblksz
+            self.verify_codeblock_style(
+                codestream.segment[2].cstyle,
+                [False, False, False, False, False, False]
+            )
+            self.assertEqual(
+                codestream.segment[2].xform,
+                glymur.core.WAVELET_XFORM_5X3_REVERSIBLE
+            )
+            self.assertEqual(
+                codestream.segment[2].precinct_size,
+                (
+                    (16, 16), (32, 32), (64, 64), (128, 128), (128, 128),
+                    (128, 128)
+                )
+            )
 
     def test_NR_ENC_Bretagne2_ppm_5_encode(self):
         """
@@ -1343,9 +1487,10 @@ class TestJp2k_write(fixtures.MetadataBase):
             self.assertEqual(codestream.segment[2].num_res, 5)  # levels
             self.assertEqual(tuple(codestream.segment[2].code_block_size),
                              (64, 64))  # cblksz
-            self.verify_codeblock_style(codestream.segment[2].cstyle,
-                                        [False, False,
-                                         False, False, False, False])
+            self.verify_codeblock_style(
+                codestream.segment[2].cstyle,
+                [False, False, False, False, False, False]
+            )
             self.assertEqual(codestream.segment[2].xform,
                              glymur.core.WAVELET_XFORM_5X3_REVERSIBLE)
             self.assertEqual(codestream.segment[2].precinct_size,
@@ -1565,10 +1710,16 @@ class TestJp2k_write(fixtures.MetadataBase):
 
             codestream = jp2.box[3].codestream
 
-            kwargs = {'rsiz': 0, 'xysiz': (2592, 1456), 'xyosiz': (0, 0),
-                      'xytsiz': (2592, 1456), 'xytosiz': (0, 0),
-                      'bitdepth': (8, 8, 8), 'signed': (False, False, False),
-                      'xyrsiz': [(1, 1, 1), (1, 1, 1)]}
+            kwargs = {
+                'rsiz': 0,
+                'xysiz': (2592, 1456),
+                'xyosiz': (0, 0),
+                'xytsiz': (2592, 1456),
+                'xytosiz': (0, 0),
+                'bitdepth': (8, 8, 8),
+                'signed': (False, False, False),
+                'xyrsiz': [(1, 1, 1), (1, 1, 1)]
+            }
             self.verifySizSegment(codestream.segment[1],
                                   glymur.codestream.SIZsegment(**kwargs))
 
@@ -1602,10 +1753,16 @@ class TestJp2k_write(fixtures.MetadataBase):
 
             codestream = j.get_codestream(header_only=False)
 
-            kwargs = {'rsiz': 0, 'xysiz': (1024, 1024), 'xyosiz': (0, 0),
-                      'xytsiz': (1024, 1024), 'xytosiz': (0, 0),
-                      'bitdepth': (16,), 'signed': (False,),
-                      'xyrsiz': [(1,), (1,)]}
+            kwargs = {
+                'rsiz': 0,
+                'xysiz': (1024, 1024),
+                'xyosiz': (0, 0),
+                'xytsiz': (1024, 1024),
+                'xytosiz': (0, 0),
+                'bitdepth': (16,),
+                'signed': (False,),
+                'xyrsiz': [(1,), (1,)]
+            }
             self.verifySizSegment(codestream.segment[1],
                                   glymur.codestream.SIZsegment(**kwargs))
 
